// Paul Houle (paulhoule.com) March 19, 2008
//
//  This file is part of par2cmdline (a PAR 2.0 compatible file verification and
//  repair tool). See http://parchive.sourceforge.net for details of PAR 2.0.
//
//  par2cmdline is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  par2cmdline is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "par2cmdline.h"

#pragma warning(disable:4731)		// inhibit VC's "ebp modified" warning
#pragma warning(disable:4799)		// inhibit VC's "missing emms" warning

unsigned PrCap = 0;
unsigned PrNum = 0;

// Sets PrCap to 1 for non-MMX processor, 2 for MMX processor.
// Sets PrNum to number of processors.
// Only needs to be called once immediately after parsing command line.

void SetPrCap(const CommandLine &commandline) {
	SYSTEM_INFO si;

	if (PrCap) return;

  _asm {
	pushfd								;try to flip "ID" bit in eflags
	pop     eax
	btc     eax,21
	push    eax
	popfd
	pushfd
	pop		ecx
	xor     ecx,eax						;bit 21 == 0 if "ID" bit changed
	xor		eax,eax
	bt      ecx,21
	jc		fset						;jmp= no CPUID, no advanced opcodes
	cpuid								;eax= highest supported CPUID function
	test	eax,eax
	jz		fset						;jmp= can't get features
	mov		eax,1						;get features
	cpuid
	xor		eax,eax
	test	edx,1						;floating point?
	jz		fset						;jmp= no, no MMX
	bt		edx,23						;MMX support?
	jnc		fset						;jmp= nope
	inc		eax							;indicate MMX
fset:
	inc		eax
	mov		PrCap,eax
fdone:
  }

	if (PrCap < 2
	 && commandline.GetNoiseLevel() > CommandLine::nlQuiet) {
		cout << "*** Warning: non-MMX processor ***" << endl;
	}

	GetSystemInfo(&si);					// ok for Win95+, all WinNT
	PrNum = si.dwNumberOfProcessors;	// fetch # of processors
}

/*----------------------------------------------------------------------------
;
; Builds the combined multiplication tables at the specified address (pMtab).
;
;---------------------------------------------------------------------------*/

void BuildMtab(unsigned *pMtab, unsigned factor, void *table) {
  _asm {
	mov		eax,factor
	movzx	ebx,ah
	movzx	eax,al
	shl		ebx,8
	shl		eax,8
	mov		edx,table
	lea		esi,[edx+eax*2]					;LL
	lea		ecx,[edx+ebx*2+(1*256*256)*2]	;HL
	lea		edi,[edx+eax*2+(1*256*256)*2]	;LH
	lea		edx,[edx+ebx*2+(2*256*256)*2]	;HH

	xor		ebx,ebx
	pxor	mm7,mm7
	mov		eax,pMtab
	align	16
LHmxlp:
	movq	mm0,[esi+ebx]
	pxor	mm0,[ecx+ebx]				;L3,L2,L1,L0
	movq	mm1,mm0
	punpcklwd mm0,mm7
	punpckhwd mm1,mm7
	movq	[eax+ebx*2],mm0
	movq	[eax+ebx*2+8],mm1

	movq	mm2,[edi+ebx]
	pxor	mm2,[edx+ebx]				;H3,H2,H1,H0
	movq	mm3,mm2
	punpcklwd mm2,mm7
	punpckhwd mm3,mm7
	movq	[eax+400h+ebx*2],mm2
	movq	[eax+400h+ebx*2+8],mm3

	add		ebx,8
	cmp		ebx,512
	jne		LHmxlp
  }
}

/*----------------------------------------------------------------------------
;
; Processes block of data a multiple of 4 bytes long.
;
; The amount of data to process (bsize) must be a non-zero multiple of 4.
;
;---------------------------------------------------------------------------*/

void DoBlock4(char *inputbuffer, char *outputbuffer,
	size_t bsize, unsigned *pMtab
) {
  _asm {
	push	ebp
	mov		ebx,bsize					;bytes to process (multiple of 4)
	mov		esi,inputbuffer				;source
	mov		edi,outputbuffer			;destination
	mov		ebp,pMtab					;combined multiplication table

lp4:
	mov		eax,[esi]
	add		esi,4
	mov		ecx,eax
	shr		eax,16
	movzx	edx,al
	shr		eax,8
	mov		edx,[ebp+edx*4]				;look up 2nd solution
	xor		edx,[ebp+400h+eax*4]
	movzx	eax,ch
	movzx	ecx,cl
	shl		edx,16
	mov		eax,[ebp+400h+eax*4]		;1st solution
	xor		eax,[ebp+ecx*4]
	xor		edx,eax						;combined solution
	xor		edx,[edi]
	mov		[edi],edx
	add		edi,4
	sub		ebx,4						;adjust src/dst count, zero if done
	jnz		lp4

	pop		ebp
  }
}

/*----------------------------------------------------------------------------
;
; Processes block of data a multiple of 8 bytes long using SIMD (mmx) opcodes.
;
; It is the caller's responsibility to insure the in/out pointers are
; optimally aligned; if not, performance is reduced significantly.
;
; The amount of data to process (bsize) must be a non-zero multiple of 8.
;
; Unfortunately this routine "reads ahead" 8 bytes beyond the true end of
; the given source block.  The caller must compensate for this by
; appropriately limiting the size of the source.
;
;---------------------------------------------------------------------------*/

void DoBlock8(char *inputbuffer, char *outputbuffer,
	size_t bsize, unsigned *pMtab
) {
  _asm {
	push	ebp
	mov		ebx,bsize					;bytes to process (multiple of 4)
	mov		esi,inputbuffer				;source
	mov		edi,outputbuffer			;destination
	mov		ebp,pMtab					;combined multiplication table

	mov		eax,[esi]					;load 1st 8 source bytes
	movd	mm4,[esi+4]

	add		esi,ebx						;point to end of input/output
	add		edi,ebx
	neg		ebx							;convert byte size to count-up

	; The below loop processes 8 bytes at a time.  It runs at approx.
	; 10.5 clocks per iteration on a Core 2 Duo.
	; That's close to optimal -- which is 10 -- since there are 10 mem
	; reads in the loop (there's only one mem read port in the
	; architecture).  The extra .5 clock appears to be a
	; cache-loading delay, since if the read addresses below are made
	; invariant it goes away.
	;
	; The 8 byte MMX loop runs 20-30% faster than a simple (non-MMX) four
	; byte loop on a Core 2 Duo.  The speedup is really only due to
	; it doing wider memory fetches (8 bytes per src/dst instead of 4).
	;
	; Some work was done to achieve U/V pipe pairing for early Pentiums,
	; but not at the expense of the Core 2 Duo performance.

	align	16
lp8:
	movzx	edx,al
	movzx	ecx,ah
	shr		eax,16
	movd	mm0,[ebp+edx*4]
	movd	mm1,[ebp+400h+ecx*4]
	movzx	edx,al
	movzx	ecx,ah
	movd	eax,mm4
	movq	mm4,[esi+ebx+8]				;read-ahead next 8 source bytes
	movd	mm2,[ebp+edx*4]
	movzx	edx,al
	movq	mm5,[edi+ebx]
	movd	mm3,[ebp+400h+ecx*4]
	movzx	ecx,ah
	shr		eax,16
	punpckldq mm0,[ebp+edx*4]
	movzx	edx,al
	punpckldq mm1,[ebp+400h+ecx*4]
	movzx	ecx,ah
	punpckldq mm2,[ebp+edx*4]
	pxor	mm1,mm0
	punpckldq mm3,[ebp+400h+ecx*4]
	pxor	mm3,mm2
	pslld	mm3,16
	movd	eax,mm4						;prepare src bytes 3-0 for next loop
	pxor	mm1,mm5
	psrlq	mm4,32						;align src bytes 7-4 for next loop
	pxor	mm1,mm3
	movq	[edi+ebx],mm1
	add		ebx,8
	jnz		lp8

	pop		ebp
  }
}

/*----------------------------------------------------------------------------
;
; Threading logic (init, release control struct) and thread function itself.
;
;---------------------------------------------------------------------------*/

// It behooves us to have more "work blocks" than threads to process them.
// If the two amounts were equal (say both were 4), and one processor out of
// 4 was delayed 1/2 the time it took to process its block, the other 3
// processors would be idle for that 1/2 block time period -- since they
// would be left with no data to process.
//
// So we try to generate lots of smaller blocks, while insuring they
// are large enough to make paying the threading overhead worthwhile.

#define MAXTH 32						// maximum threads (processors)
#define MINPT 1024						// minimum size per thread

typedef struct {						// RS threading control struct
	unsigned init;						// non-zero after struct initialized
	unsigned nThreads;					// created threads (# processors - 1)
				// volatile fields (may be shared among multiple threads)
	unsigned * volatile pMtab;			// ptr to true start of Mtab
	long volatile bltstart;				// blocks left to start
	long volatile bltfinish;			// blocks left to finish
	int volatile bsize;					// size of non-last blocks (mult of 8)
	int volatile bsizelast;				// size of     last block  (mult of 8)
	char * volatile inbuflast;			// start of last block in inputbuffer
	char * volatile outbuflast;			// start of last block in outputbuffer
				// kernel objects
	HANDLE sem;							// semaphore used to dispatch threads
	HANDLE ev;							// event used to sense completion
	HANDLE tH[MAXTH];					// vector of thread handles
} RSTH;
RSTH RSth = {0};

void RelRSthrd(RSTH * const pRSth) {	// release RS threading resources
	int i;
	
	if (pRSth->nThreads) {				// if thread(s) were created,
		pRSth->bltstart = -9999;		// command all threads to exit
		ReleaseSemaphore(pRSth->sem, pRSth->nThreads, NULL);
		WaitForMultipleObjects(			// wait for all threads to exit
			pRSth->nThreads, pRSth->tH, TRUE, 5000);
		for (i = 0; i < pRSth->nThreads; i++) {
			CloseHandle(pRSth->tH[i]);	// release thread handles
		}
	}
	if (pRSth->ev) CloseHandle(pRSth->ev);	// close event & semaphore handles
	if (pRSth->sem) CloseHandle(pRSth->sem);
}

// This routine processes blocks specified in RSth.  It continues as long as
// it can get a positive number from decrementing RSth->bltstart (meaning
// there's another block-left-to-start), AND after processing such a block,
// it doesn't get a zero from decrementing RSth-bltfinish (meaning there's
// more blocks left-to-finish).
//
//		Returns <0 if exit due to blocks-left-to-start went negative;
//			return value is blocks-left-to-start.
//
//		Returns =0 if exit due to blocks-left-to-finish hit zero --
//			important: that indicates all work done for all threads
//			(only one zero will be returned among all callers).
//
// Note it is possible -- though unlikely -- that successive loops below will
// operate on completely different input buffers (eg inbuflast & outbuflast
// change), without this thread ever going to sleep and being awakened.
// That's OK.

long DoBlocks(RSTH * const pRSth) {		// process block(s)
	long bleft;
	unsigned boff;

		// pRSth->bltstart is blocks left to process.
		// bltstart synchronizes access to the other block fields -- if >= 0,
		// no other thread (ie the block dispatch thread) will be writing
		// to the other block fields.  However, the block fields may change
		// between successive positve bltstart values.
		// Below:
		//		If bleft <  0, there's no more blocks left to process.
		//		If bleft == 0, we'll be processing the last block.

		do {
			bleft = InterlockedDecrement(&pRSth->bltstart);
			if (bleft < 0) break;
			boff = bleft * pRSth->bsize;
			DoBlock8(					// process a work block
				pRSth->inbuflast - boff, pRSth->outbuflast - boff,
				bleft ? pRSth->bsize : pRSth->bsizelast, pRSth->pMtab
			);
			bleft = InterlockedDecrement(&pRSth->bltfinish);
		} while (bleft);

	return bleft;
}

// Note: the main thread function (below) and the RS work function it calls
// do not use any C run-time functions; hence, we can use CreateThread
// instead of beginthreadex.

DWORD WINAPI ThreadFunc(LPVOID targ) {	// main thread function
	RSTH * const pRSth = (RSTH * const) targ;
	long dbret;

	do {
		WaitForSingleObject(pRSth->sem, INFINITE); // await triggering
		dbret = DoBlocks(pRSth);		// try to process a block
		if (dbret == 0) {				// if we sensed end of all work to do,
			SetEvent(pRSth->ev);		// need to wake up the main thread
		}
	} while (dbret > -9999);			// operate till commanded to exit

	return 0;
}

void DoRelRSthrd(void) {				// needed for atexit() in InitRSthrd
	RelRSthrd(&RSth);
}

// Initializes the threading fields of the thread control struct, creating
// one thread per available processor.
// If multi-processor operation is not possible, sets RSth.nThreads == 0.

void InitRSthrd(RSTH * const pRSth) {	// init RS threading resources
	int i, j;
	long volatile iitest = 99;			// for InterlockedIncrement test

	memset(pRSth, 0, sizeof(*pRSth));	// init thread ctl struct to all zero
	pRSth->init++;						// only run this init code once

	i = (int) PrNum;					// # of processors
	--i;								// create 1 thread per extra processor
	for (;;) {
		if (i <= 0) break;				// need at least two processors
		// need an OS w/modern InterlockedIncrement (> Win 95 & NT 3.51)
		if (InterlockedIncrement(&iitest) != 100) break;

		i = (i <= MAXTH) ? i : MAXTH;	// limit to our max processors
		pRSth->sem = CreateSemaphore(	// semaphore to dispatch threads
			NULL, 0, i * 2, NULL);		//	(non-signaled)
		if (!pRSth->sem) break;
		pRSth->ev = CreateEvent(		// event to sense end of processing
			NULL, FALSE, FALSE, NULL);	//	(non-signaled, auto-reset)
		if (!pRSth->ev) break;
		for (j = 0; j < i; j++) {		// create threads for extra processors
			DWORD ThreadID;
			pRSth->tH[j] = CreateThread(
				NULL, 2048, ThreadFunc, &RSth, 0, &ThreadID);
			if (!pRSth->tH[j]) break;
			pRSth->nThreads++;
		}
		break;
	}
	if (iitest == 100) {				// if we tried to create threads,
		if (i == pRSth->nThreads) {		//	and were entirely successful,
			atexit(DoRelRSthrd);		// install at-exit cleanup routine
		} else {						// else we tried and failed,
			RelRSthrd(pRSth);			// kill anything halfway done
			pRSth->nThreads = 0;		// disable threading
		}
	}
}

/*----------------------------------------------------------------------------
;
; Paul Houle (paulhoule.com) March 19, 2008
; The below routine is the gist of the changes.
; It runs the assembly code to handle the main CPU work of par2 creation.
;
; Note: the entry bsize is required to be a multiple of four.
;
;---------------------------------------------------------------------------*/

void AsmRSPrc(size_t bsize, char *inputbuffer,
  char *outputbuffer, unsigned factor, void *table) {

	char Mtab_space[256*2*4+15];		// mult tables (+ extra for alignment)
	unsigned *pMtab;					// ptr to true start of Mtab

	pMtab = (unsigned *)				// para align true Mtab start addr
		((unsigned)(Mtab_space + 15) & ~15);

	BuildMtab(pMtab, factor, table);	// build Mtab (combined mult. tables)

	for (;;) {

		// Handle tiny blocks with straightforward 4-byte routine for two
		// reasons: (1) the 8-byte logic requires a minimum blocksize to
		// operate -- currently 24; (2) for tiny blocks it's faster to use
		// the 4-byte routine since it has so much less setup overhead.
		//
		// (the "40" number below was determined empirically).

		if (bsize <= 40) {				// if too small to get fancy with,
			if (bsize) {				// if something at all,
				DoBlock4(inputbuffer, outputbuffer, bsize, pMtab);
			}
			break;
		}

		// Here we'll try 8-byte blocks, and perhaps multi-threading as well.
		//
		// inbuf & outbuf are always relatively 8-byte aligned, but the caller
		// operates in dword sizes so they may not be absolutely 8-byte
		// aligned.  Misalignment slows things down quite a bit (10%); so, if
		// addresses misaligned, process a dword beforehand to align them.

		if ((int) outputbuffer & 4) {	// if out misaligned (in will be too),
			DoBlock4(inputbuffer, outputbuffer, 4, pMtab);
			inputbuffer += 4;
			outputbuffer += 4;
			bsize -= 4;
		}

		// We don't bother multi-threading unless there's a significant
		// amount (MINPT) of data for at least two threads -- the overhead
		// to setup/teardown threading doesn't make it worthwhile.

		if (bsize >= MINPT * 2) {	// if worthwhile to try threading,
			if (!RSth.init) {		// if haven't initialized threading,
				InitRSthrd(&RSth);	// try to create 1 thread per extra core
			}
		}

		if (bsize < MINPT * 2			// if too small to bother threading,
		 || !RSth.nThreads) {			//	or no threads could be created,
			size_t bsz = bsize & ~4;	// size must be a multiple of 8
			bsz -= 8;					// compensate for read-ahead
			DoBlock8(					// process large chunk 8-bytes a shot
				inputbuffer, outputbuffer, bsz, pMtab);
			DoBlock4(					// process residual 8-12 bytes
				inputbuffer + bsz, outputbuffer + bsz, bsize - bsz, pMtab);
			break;
		}

		// At this point we've decided to use multiple threads (and we have
		// them).  The overhead is relatively the same whether we start
		// 1 or 31.  If we have only (MINPT * 2) bytes it might get spread
		// across lots of threads (for a blocksize much smaller than MINPT),
		// but not much should be lost if that happens.

		{
			RSTH * const pRSth = &RSth;
			unsigned btmp;

			pRSth->pMtab = pMtab;		// set pMtab for all threads

			// Set btmp to the # blocks we'll divide the source into.
			// We try to divide the source up into approximately MINPT blocks.
			// But we also keep the # blocks a multiple of the # of threads,
			// to (hopefully) keep all the cores at 100% utilization.
			//
			// Note: you see +1 on the thread count below since nThreads
			// is the CREATED threads -- but we do work in this thread too --
			// so the total number of work threads is nThreads + 1.

			btmp = bsize / MINPT;		// approx. # of blocks we'll shoot for
			btmp -=						// round to a multiple of # threads
				(btmp % (pRSth->nThreads + 1));
			if (!btmp) btmp = pRSth->nThreads + 1;

			// Set number of blocks that must be started and finished (this
			// is the total number of blocks).

			RSth.bltstart = btmp;
			RSth.bltfinish = btmp;

			// Here we set the block sizes -- there are two.  bsize is the
			// size of every block except the last.  bsizelast is the size
			// of the last block.  Both need to be a non-zero multiple of 8,
			// since the threading work routines require that.
			//
			// Note the last block is reduced by 8 to compensate for the
			// read-ahead done in the Block8 routine.

			pRSth->bsize = ((bsize / btmp) + 7) & ~7;
			btmp = (btmp - 1) * pRSth->bsize;
			pRSth->bsizelast = ((bsize - btmp) & ~7) - 8;

			pRSth->inbuflast = inputbuffer + btmp;
			pRSth->outbuflast = outputbuffer + btmp;
			ReleaseSemaphore(pRSth->sem, pRSth->nThreads, NULL);

			if (DoBlocks(pRSth) < 0) {	// process block(s) in this thread
				WaitForSingleObject(pRSth->ev, INFINITE);
			}
			btmp =						// btmp= bytes processed by threads
				(pRSth->inbuflast + pRSth->bsizelast) - inputbuffer;
			DoBlock4(					// process residual 8-12 bytes
				inputbuffer + btmp, outputbuffer + btmp, bsize - btmp, pMtab);
		}

		break;
    }

	_asm emms
}
